/**
 * (C) 2014 Jan Holst Jensen, jan@biochemfusion.com.
 *
 * This C source code unit (pypl.c) is released under a BSD-style license:
 *
 * Copyright (C) 2014 Biochemfusion Holding ApS (http://www.biochemfusion.com)
 * All rights reserved.
 *
 * Redistribution and use for any purpose in source and binary forms, with or
 * without modification, are permitted, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY Biochemfusion Holding ApS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Biochemfusion Holding ApS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/

#include <Python.h>

#ifdef _WIN32
	#define ON_WINDOWS
#endif

#ifndef ON_WINDOWS
	#include <dlfcn.h>
#endif

#define RETURN_CODE_SUCCESS 0
#define RETURN_CODE_ERROR 1
#define MAX_ERRMSG_LEN 512

PyObject* pymain      = NULL;
PyObject* global_dict = NULL;

char err_msg[MAX_ERRMSG_LEN + 1] = "";
const char* OK_NO_RESULT = "(OK - no result)";
const char* ERROR_NO_RESULT = "(ERROR - no result)";
const char* NULL_RESULT = "(null)";

#ifndef ON_WINDOWS
// The Python shared library will be preloaded with special flags to avoid
// import errors when using numpy. See also
//   http://stackoverflow.com/questions/20849755/python-shared-libraries-rtld-global-segfault
//   http://theneural.wordpress.com/2012/02/13/calling-python-from-matlab/
int is_python_preloaded = 0;
// Change LIBPYTHON_PATH to match your Python version. This is for the default
// Python 2.6 on CentOS 6.5.
const char* LIBPYTHON_PATH = "libpython2.6.so.1.0";
#endif

inline void copy_string_to_orabuf(const char* s, char* buf, int* buf_maxlen)
{
  strncpy(buf, s, *buf_maxlen);
  buf[*buf_maxlen] = 0; // Necessary, or Oracle may complain about a length mismatch problem.
}

#ifdef __cplusplus
  extern "C" {
#endif

const char* latest_err_msg()
{
  return err_msg;
}

int reset()
{
#ifndef ON_WINDOWS
  // This is only relevant on *NIX type systems.
  if (!is_python_preloaded) {
    dlopen(LIBPYTHON_PATH, RTLD_NOW|RTLD_GLOBAL);
    const char* preload_err = dlerror();
    if (preload_err) {
      strncpy(err_msg, preload_err, MAX_ERRMSG_LEN);
      err_msg[MAX_ERRMSG_LEN] = 0;
      return RETURN_CODE_ERROR;
    }
    is_python_preloaded = 1;
  }
#endif

  // Release runtime if initialized.
  if (global_dict != NULL) {
    // 'global_dict' and 'pymain' are borrowed references. Do not decr their refcounts.
    global_dict = NULL;
    pymain      = NULL;
    Py_Finalize();
  }

  Py_Initialize();

  pymain      = PyImport_AddModule("__main__");
  global_dict = PyModule_GetDict(pymain);

  if (global_dict != NULL)
    return RETURN_CODE_SUCCESS;
  else {
    strcpy(err_msg, "Failed to initialize Python runtime.");
    return RETURN_CODE_ERROR;
  }
}

int run_script(const char* script_text, const char* result_varname, char* result, int* result_maxlen)
{
  if (global_dict == NULL) {
    if (reset() != RETURN_CODE_SUCCESS)
      return RETURN_CODE_ERROR;
  }

  PyObject* script_result = PyRun_String(script_text, Py_file_input, global_dict, global_dict);
  if (script_result == NULL) {
    PyObject *ptype, *pvalue, *ptraceback;
    PyErr_Fetch(&ptype, &pvalue, &ptraceback);
    PyObject *value_as_str = PyObject_Str(pvalue);
    char* error_string = PyString_AsString(value_as_str);

    if (error_string == NULL)
      strcpy(err_msg, "Unspecified exception from Python.");
    else
      strncpy(err_msg, error_string, MAX_ERRMSG_LEN);
    // If error message is longer than 'err_msg' buffer strncpy() won't guarantee that
    // 'err_msg' now is null-terminated. Make sure it is.
    err_msg[MAX_ERRMSG_LEN] = 0;

    copy_string_to_orabuf(ERROR_NO_RESULT, result, result_maxlen);
    return RETURN_CODE_ERROR;
  }

  // If no result is requested from Oracle a " " string will be passed. Allowing passing NULL
  // directly complicates the call interface.
  if (result_varname == NULL || (*result_varname == ' ' && *(result_varname + 1) == 0))
    copy_string_to_orabuf(OK_NO_RESULT, result, result_maxlen);
  else {
    PyObject* py_result = PyDict_GetItemString(global_dict, result_varname);
    if (py_result == NULL)
      copy_string_to_orabuf(NULL_RESULT, result, result_maxlen);
    else
      copy_string_to_orabuf(PyString_AsString(py_result), result, result_maxlen);
  }

  return RETURN_CODE_SUCCESS;
}
#ifdef __cplusplus
  } // end of 'extern "C"' block.
#endif

#ifdef CONSOLE_TEST
/**
  DEBUG: Set CONSOLE_TEST and recompile to run a test as a console application.
**/

int main(int argc, char *argv[])
{
  Py_SetProgramName(argv[0]);  // Optional but recommended.

  char result[30000];
  int result_maxlen = 30000;
  int rc;

  rc = reset();
  if (rc != RETURN_CODE_SUCCESS)
    printf("ERROR: %s\n", err_msg);

  const char* pythonScript1 = 
    "def multiply(x, y):\n"
    "  return x * y\n";

  rc = run_script(pythonScript1, NULL, &result[0], &result_maxlen);
  if (rc != RETURN_CODE_SUCCESS)
    printf("ERROR: %s\n", err_msg);

  const char* pythonScript2 = 
    "x = 2\n"
    "y = 6\n"
    "result = str(multiply(x, y))\n";

  rc = run_script(pythonScript2, "result", &result[0], &result_maxlen);
  if (rc != RETURN_CODE_SUCCESS)
    printf("ERROR: %s\n", err_msg);
  else
    printf("x * y = %s\n", result);

  // This should fail.
  const char* pythonScript3 = 
    "x = '2'\n"
    "y = '6'\n"
    "result = str(multiply(x, y))\n";

  rc = run_script(pythonScript3, "result", &result[0], &result_maxlen);
  if (rc != RETURN_CODE_SUCCESS)
    printf("ERROR: %s\n", err_msg);
  else
    printf("x * y = %s\n", result);

#undef TEST_RESET
#ifdef TEST_RESET
  // Show how we leak a bit of memory every time reset() is called.
  int i;
  for (i = 0; i < 100000; i++) {
    reset();
    rc = run_script("result = str(34)", "result", &result[0], &result_maxlen);
    if (rc != RETURN_CODE_SUCCESS) {
      printf("FATAL ERROR: %s\n", err_msg);
      return 1;
    }
  }
#endif

  // If the global dictionary is not properly updated between calls to PyRun_String()
  // then the following script will fail with "global name 'f1' is not defined". This
  // would happen in the initial release 2014-02-27.
  rc = run_script(
    "import datetime\n"
    "\n"
    "def f1():\n"
    "  return datetime.datetime.now()\n"
    "\n"
    "def f2():\n"
    "  return f1()\n"
    "\n"
    "result = str(f2())\n"
  , "result", &result[0], &result_maxlen);
  if (rc != RETURN_CODE_SUCCESS)
    printf("ERROR: %s\n", err_msg);
  else
    printf("f2() = %s\n", result);

  printf("OK - all done.\n");
  Py_Finalize();
  return 0;
}
#endif
